/*
* Copyright (c) 2017 elementary LLC
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either
* version 3 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*
* Authored by: Jeremy Wootten <jeremy@elementaryos.org>
*/

void add_gof_directory_async_tests () {
    /* loading */
    Test.add_func ("/GOFDirectoryAsync/load_non_existent_local", () => {
        run_load_folder_test (load_non_existent_local_test);
    });
    Test.add_func ("/GOFDirectoryAsync/load_empty_local", () => {
        run_load_folder_test (load_empty_local_test);
    });
    Test.add_func ("/GOFDirectoryAsync/load_populated_local", () => {
        run_load_folder_test (load_populated_local_test);
    });
    Test.add_func ("/GOFDirectoryAsync/load_cached_local", () => {
        run_load_folder_test (load_cached_local_test);
    });
    Test.add_func ("/GOFDirectoryAsync/reload_populated_local", () => {
        run_load_folder_test (reload_populated_local_test);
    });
}

delegate GOF.Directory.Async LoadFolderTest (string path, MainLoop loop);
void run_load_folder_test (LoadFolderTest test) {
    var loop = new GLib.MainLoop ();
    string test_dir_path = "/tmp/marlin-test-" + get_real_time ().to_string ();

    GOF.Directory.Async dir = test (test_dir_path, loop);

    dir.init ();
    loop.run ();

    /* Tear down test folder*/
    tear_down_folder (test_dir_path);
}

/*** Test functions ***/
GOF.Directory.Async load_non_existent_local_test (string test_dir_path, MainLoop loop) {
    GLib.File gfile = GLib.File.new_for_commandline_arg (test_dir_path);
    assert (!gfile.query_exists (null));

    GOF.Directory.Async dir = GOF.Directory.Async.from_gfile (gfile);
    dir.done_loading.connect (() => {
        assert (dir.files_count == 0);
        assert (!dir.can_load);
        Test.assert_expected_messages ();
        loop.quit ();
    });

    Test.expect_message (null, GLib.LogLevelFlags.LEVEL_WARNING,"*info*");
    Test.expect_message (null, GLib.LogLevelFlags.LEVEL_WARNING,"*cannot load*");

    return dir;
}

GOF.Directory.Async load_empty_local_test (string test_dir_path, MainLoop loop) {
    var dir = setup_temp_async (test_dir_path, 0);

    dir.done_loading.connect (() => {
        assert (dir.files_count == 0);
        assert (dir.can_load);
        loop.quit ();
    });

    return dir;
}

GOF.Directory.Async load_populated_local_test (string test_dir_path, MainLoop loop) {
    uint n_files = 5;
    uint file_loaded_signal_count = 0;

    var dir = setup_temp_async (test_dir_path, n_files);

    assert (dir.ref_count == 1); //First ref replaced by toggle ref;

    dir.file_loaded.connect (() => {
        file_loaded_signal_count++;
    });

    dir.done_loading.connect (() => {
        assert (dir.files_count == n_files);
        assert (dir.can_load);
        assert (dir.state == GOF.Directory.Async.State.LOADED);
        assert (file_loaded_signal_count == n_files);

        Test.assert_expected_messages ();

        loop.quit ();
    });

    Test.expect_message (null, GLib.LogLevelFlags.LEVEL_DEBUG,"*(Re)loading*");
    return dir;
}

GOF.Directory.Async load_cached_local_test (string test_dir_path, MainLoop loop) {
    uint n_files = 5;
    bool first_load = true;
    uint file_loaded_signal_count = 0;

    var dir = setup_temp_async (test_dir_path, n_files);

    dir.done_loading.connect (() => {
        if (first_load) {
            first_load = false;
            dir.file_loaded.connect (() => {
                file_loaded_signal_count++;
            });

            Test.expect_message (null, GLib.LogLevelFlags.LEVEL_DEBUG,"*cached*");
            dir.init ();
        } else {
            assert (dir.files_count == n_files);
            assert (dir.can_load);
            assert (dir.state == GOF.Directory.Async.State.LOADED);
            assert (file_loaded_signal_count == n_files);
            Test.assert_expected_messages ();

            loop.quit ();
        }
    });
    return dir;
}

GOF.Directory.Async reload_populated_local_test (string test_dir_path, MainLoop loop) {
    uint n_files = 50;
    uint n_loads = 5; /* Number of times to reload the directory */
    uint loads = 0;
    uint ref_count_before_reload = 0;
    string tmp_pth = get_text_template_path ();

    var dir = setup_temp_async (test_dir_path, n_files, "txt", tmp_pth);

    dir.done_loading.connect (() => {
        if (loads == 0) {
            ref_count_before_reload = dir.ref_count;
            
        }
        if (loads < n_loads) {
            loads++;
            dir.cancel ();
            dir.reload ();
        } else {
            assert (dir.files_count == n_files);
            assert (dir.can_load);
            assert (dir.state == GOF.Directory.Async.State.LOADED);
            assert (dir.ref_count == ref_count_before_reload);

            Test.assert_expected_messages ();

            tear_down_file (tmp_pth);

            /* Test for problem with toggle ref after reloading (lp:1665620) */
            dir.cancel ();
            dir = null;

            loop.quit ();
        }
    });

    Test.expect_message (null, GLib.LogLevelFlags.LEVEL_DEBUG,"*(Re)loading*");
    return dir;
}

/*** Helper functions ***/
GOF.Directory.Async setup_temp_async (string path, uint n_files, string? extension = null, string? path_to_template = null) {
    assert (extension == null || extension.length > 0 || extension.length < 5);

    Posix.system ("mkdir " + path);
    string extn = "";

    if (extension != null) {
        extn = "." + extension;
    }

    for (int i = 0; i < n_files; i++) {
        string pth = path + Path.DIR_SEPARATOR_S + i.to_string () + extn;
        if (path_to_template == null) {
            /* create empty files */
            Posix.system ("touch " + pth);
        } else {
            Posix.system ("cp --no-dereference --no-clobber " + path_to_template + " " + pth);
        }
    }

    GLib.File gfile = GLib.File.new_for_commandline_arg (path);
    assert (gfile.query_exists (null));

    GOF.Directory.Async dir = GOF.Directory.Async.from_gfile (gfile);
    assert (dir != null);
    return dir;
}

string get_text_template_path () {
    string test_template_path = "/tmp/marlin-template-" + get_real_time ().to_string () + "txt";
    Posix.system ("env > " + test_template_path);
    return test_template_path;
}

void tear_down_folder (string path) {
    Posix.system ("rm -rf " + path);
}

void tear_down_file (string path) {
    Posix.system ("rm -f " + path);
}

int main (string[] args) {
    Test.init (ref args);

    add_gof_directory_async_tests ();
    return Test.run ();
}
